BEGIN TRANSACTION;

--
-- Account table.
--
-- Record the mapping between a UID/GID and a user/group name.
--
--  aid     : acoount table auto increment id
--  type    : account type, values [0|1] means [user|group]
--  id      : account ID [UID|GID], depends on the "type" field
--  name    : account name
--

CREATE TABLE account (
  aid  INTEGER  primary key AUTOINCREMENT,
  type INTEGER  NOT NULL,
  id   INTEGER  NOT NULL,
  name TEXT     NOT NULL,
  UNIQUE (type, id) on CONFLICT REPLACE
);
CREATE INDEX a_id_idx ON account(id);

--
-- Blocked address infomation table.
--
-- Record the blocked address of information.
--
--  bai_id        : Blocked address infomation table auto increment id
--  policy_name   : the policy name
--  block_address : the address need to blocked
--

CREATE TABLE block_address_info (
  bai_id        INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  block_address TEXT  NOT NULL,
  policy_name   TEXT  NOT NULL
);

--
-- Blocked address target table.
--
-- Record the blocked address of target(users/groups).
--
--  bat_id :  Blocked address target table auto increment id
--  bai_id : mapping to block_address_info's bai_id
--  aid : mapping to account's aid
--

CREATE TABLE block_address_target (
  bat_id        INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  bai_id        INTEGER NOT NULL,
  aid           INTEGER NOT NULL,
  UNIQUE (bai_id, aid),
  FOREIGN KEY (aid) REFERENCES account (aid),
  FOREIGN KEY (bai_id) REFERENCES block_address_info (bai_id) ON DELETE CASCADE
);

--
-- Connectivity table.
--
-- Record when did a user start/stop using a service.
--
--  cid   :  Connectivity table auto increment id
--  aid   : a user account ID, mapping to account's aid
--  pid   : protocol id, mapping to protocol's pid
--  signature : service-dependent signature, dependents on the "protocol" field
--          (1) VPN service
--            the VPN IP assigned to the user
--          (2) Web portal
--  ip_from   : the IP address where the user connected from
--  time_start  : when did the user start using the service
--  time_stop : when did the user stop using the service,
--          equal to INT64_MAX if the user is still using this service
--  status    : record status code, 0 means normal, others are abnormal, default to 0
--

CREATE TABLE connectivity (
  cid         INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  aid         INTEGER NOT NULL,
  username    TEXT    NOT NULL,
  pid         INTEGER NOT NULL,
  signature   TEXT    NOT NULL,
  ip_from     TEXT    NOT NULL,
  time_start  INTEGER NOT NULL,
  time_stop   INTEGER NOT NULL,
  status      INTEGER NOT NULL  DEFAULT 0,
  UNIQUE (pid, signature, time_stop) on CONFLICT ABORT,
  FOREIGN KEY (aid) REFERENCES account (aid),
  FOREIGN KEY (pid) REFERENCES protocol (pid)
) ;
CREATE INDEX c_time_idx ON connectivity(time_stop, time_start);
CREATE INDEX c_id_idx ON connectivity(aid, pid);

--
-- Connectivity of webportal table.
--
-- Record the detail of the user surf internet via webvpn
--
--  cw_id   :  Connectivity webportal table auto increment id
--  cid   : a connection ID, mapping to aonnectivity's aid
--  time_start  : when did the user visit the target url
--  time_stop : when did the user exit the target url
--  target    :  the web site which the user visited
--

CREATE TABLE connectivity_webportal (
  cw_id         INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  cid           INTEGER NOT NULL,
  time_start    INTEGER NOT NULL,
  time_stop     INTEGER NOT NULL,
  target        TEXT NOT NULL,
  FOREIGN KEY (cid) REFERENCES connectivity (cid) ON DELETE CASCADE
) ;
CREATE INDEX cw_time_idx ON connectivity_webportal(time_stop, time_start);

--
-- default portal permission table.
--
-- Record the user who has the permission to access default portal
--
-- dpp_id :  default portal permission table auto increment id
--  aid   : a user account ID, mapping to account's aid
--

CREATE TABLE default_portal_permission (
  dpp_id      INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  aid         INTEGER NOT NULL,
  UNIQUE (aid),
  FOREIGN KEY (aid) REFERENCES account (aid)
) ;

--
-- Portal info table.
--
--  pi_id : id of a portal
--  entry   : portal entry name which will be a random value or set by user
--  name    : portal name for displaying on UI
--  ph_id   : host id mapping to portal_host
--  path    : the path of portal
--  display   : show in web portal UI or not
--  alias   : user set alias or not
--

CREATE TABLE portal_info (
  pi_id          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  entry          TEXT NOT NULL,
  name           TEXT NOT NULL,
  ph_id          INTEGER NOT NULL,
  path           TEXT    NOT NULL,
  display        NUMERIC NOT NULL,
  alias          NUMERIC NOT NULL,
  UNIQUE (entry),
  UNIQUE (ph_id, path),
  FOREIGN KEY (ph_id) REFERENCES portal_host (ph_id)
) ;
CREATE INDEX pi_display_idx ON portal_info(display);

--
-- Portal host table.
--
--  ph_id     : serial id of a portal's host
--  entry   : portal entry name which will be a random value or set by user
--  host    : the host of target address($domain:$port)
--

CREATE TABLE portal_host (
  ph_id    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  entry   TEXT    NOT NULL,
  host    TEXT    NOT NULL,
  UNIQUE (entry)
);

--
-- User portals table.
--
--  up_id : id of a user portal
--  aid   : a user account ID, mapping to account's aid
--  portal_id : portal id map to portal_info
--

CREATE TABLE user_portals (
  up_id      INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  aid        INTEGER NOT NULL,
  pi_id      INTEGER NOT NULL,
  UNIQUE (aid, pi_id),
  FOREIGN KEY (pi_id) REFERENCES portal_info (pi_id) ON DELETE CASCADE,
  FOREIGN KEY (aid) REFERENCES account (aid)
) ;
CREATE INDEX up_id_idx ON user_portals(aid);

--
-- Webportal users' favorite sites.
--
-- This table records every user's favorite sites, which can be seen or
-- deleted by the own user.
--
--  uf_id : id of a user favorite
--  aid   : a user account ID, mapping to account's aid
--  target      : target address which user wants to go
--  name      : title or nick name of website
--
CREATE TABLE user_favorite (
  uf_id          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  aid            INTEGER NOT NULL,
  target         TEXT NOT NULL,
  name           TEXT NOT NULL,
  UNIQUE (aid, target),
  FOREIGN KEY (aid) REFERENCES account (aid)
) ;
CREATE INDEX uf_id_idx ON user_favorite(aid);

--
-- Webportal users' history.
--
-- This table records every user's history, which can be seen or
-- deleted by the own user.
-- If you want to see complete user history that won't be modified
-- by users, please refer to connectivity table
--
--  uh_id : id of a user history
--  aid   : a user account ID, mapping to account's aid
--  entry   : portal entry name which will be a random value within reversed proxy
--  target    : target address which user wants to go
--  time    : log timestamp
--  web_title : title of website user accessing
--  web_favicon : icon of website user accessing
--

CREATE TABLE user_history (
  uh_id          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  aid            INTEGER NOT NULL,
  entry          TEXT NOT NULL,
  target         TEXT NOT NULL,
  time           INTEGER NOT NULL,
  web_title      TEXT,
  web_favicon    TEXT,
  UNIQUE (aid, entry),
  FOREIGN KEY (aid) REFERENCES account (aid)
) ;
CREATE INDEX uh_id_idx ON user_history(aid);
CREATE INDEX uh_time_idx ON user_history(time);

--
-- protocol table.
--
-- Record all available protocol.
--
-- pid     : protocol ID (an auto increment number)
-- name     : protocol name
--
CREATE TABLE protocol (
  pid   INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  name  TEXT NOT NULL,
  UNIQUE (name)
) ;

INSERT INTO protocol (pid, name) VALUES (0, 'all');
INSERT INTO protocol (name) VALUES ('vpnplus-l2tp');
INSERT INTO protocol (name) VALUES ('vpnplus-openvpn');
INSERT INTO protocol (name) VALUES ('vpnplus-pptp');
INSERT INTO protocol (name) VALUES ('vpnplus-ssl');
INSERT INTO protocol (name) VALUES ('vpnplus-sstp');
INSERT INTO protocol (name) VALUES ('web-portal');

--
-- Admin log type table.
--
-- Define types for admin log.
--
-- tid     : log type ID (an auto increment number)
-- name     : log type name
--
CREATE TABLE admin_log_type (
  alt_id   INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  name  TEXT NOT NULL,
  UNIQUE (name)
) ;

INSERT INTO admin_log_type (alt_id, name) VALUES (0, 'Misc');
INSERT INTO admin_log_type (name) VALUES ('Privilege');
INSERT INTO admin_log_type (name) VALUES ('Speed Limit');
INSERT INTO admin_log_type (name) VALUES ('Block List');
INSERT INTO admin_log_type (name) VALUES ('Service Change');
INSERT INTO admin_log_type (name) VALUES ('Object');
INSERT INTO admin_log_type (name) VALUES ('License');
INSERT INTO admin_log_type (name) VALUES ('Log');
INSERT INTO admin_log_type (name) VALUES ('Connection');

--
-- Traffic data table.
--
-- Record the traffic flow when did a user using a service.
--
--  td_id : id of a traffic data (autoincrement)
--  cid   : a connection ID, mapping to aonnectivity's aid
--  period  : time interval
--  upload    : the upload traffic where the user use
--  download  : the download traffic where the user use
--  time    : log timestamp
--

CREATE TABLE traffic_data (
  td_id    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  cid      INTEGER NOT NULL,
  period   INTEGER NOT NULL,
  upload   INTEGER NOT NULL,
  download INTEGER NOT NULL,
  time     INTEGER NOT NULL,
  FOREIGN KEY (cid) REFERENCES connectivity (cid) ON DELETE CASCADE
) ;
CREATE INDEX td_time_idx ON traffic_data(time);
CREATE INDEX td_cid_idx ON traffic_data(cid);

--
-- Service's traffic table.
--
-- Record every mins and statistic each service uploaded and downloaded
--
--  tpd_id : id of a traffic data (autoincrement)
--  pid      : protocol id
--  period  : time interval
--  upload    : the upload traffic where the service used
--  download  : the download traffic where the service used
--  online_user   : the among of online user in the timestamp
--  time    : log timestamp
--
CREATE TABLE traffic_protocol_data (
  tpd_id      INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  pid         INTEGER NOT NULL,
  period      INTEGER NOT NULL,
  upload      INTEGER NOT NULL,
  download    INTEGER NOT NULL,
  online_user INTEGER NOT NULL,
  time        INTEGER NOT NULL,
  FOREIGN KEY (pid) REFERENCES protocol (pid)
) ;
CREATE INDEX tpd_time_idx ON traffic_protocol_data(time);


--
-- Traffic infomation table.
--
-- Record the traffic of information.
--
--  ti_id  : id of a traffic info (autoincrement)
--  pid   : protocol id
--  speed_limit : speed limit of the policy
--  policy_name : the policy name
--

CREATE TABLE traffic_info (
  ti_id       INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  pid         INTEGER NOT NULL,
  policy_name TEXT NOT NULL,
  speed_limit INTEGER NOT NULL,
  FOREIGN KEY (pid) REFERENCES protocol (pid)
) ;

--
-- Traffic target table.
--
--  Record the traffic of target(users/groups).
--
--  tt_id  : id of a traffic target (autoincrement)
--  ti_id :  mapping to traffic_info's ti_id
--  aid   : a user account ID, mapping to account's aid
--

CREATE TABLE traffic_target (
  tt_id        INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  ti_id        INTEGER NOT NULL,
  aid          INTEGER NOT NULL,
  UNIQUE (ti_id, aid),
  FOREIGN KEY (aid) REFERENCES account (aid),
  FOREIGN KEY (ti_id) REFERENCES traffic_info (ti_id) ON DELETE CASCADE
) ;
CREATE INDEX tt_id_idx ON traffic_target(ti_id);

--
-- admin event log table.
--
-- Record all logs generated by admin.
--
--  al_id : id of a admin log (autoincrement)
--  aid   : a user account ID, mapping to account's aid
--  pid      : protocol id
--  message  : log message
--  level   : log level
--  time    : log timestamp
--

CREATE TABLE admin_log (
  al_id       INTEGER PRIMARY KEY AUTOINCREMENT,
  aid         INTEGER NOT NULL,
  alt_id      INTEGER NOT NULL,
  level       INTEGER NOT NULL,
  message     TEXT NOT NULL,
  time        INTEGER NOT NULL,
  FOREIGN KEY (aid) REFERENCES account (aid),
  FOREIGN KEY (alt_id) REFERENCES admin_log_type (alt_id)
);
CREATE INDEX al_time_idx ON admin_log(time);

--
-- user event log table.
--
-- Record all connection status change of user.
--
--  ul_id : id of a user log (autoincrement)
--  aid   : a user account ID, mapping to account's aid
--  pid      : protocol id
--  message  : log message
--  level   : log level
--  time    : log timestamp
--

CREATE TABLE user_log (
  ul_id       INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  aid         INTEGER NOT NULL,
  pid         INTEGER NOT NULL,
  level       INTEGER NOT NULL,
  message     TEXT NOT NULL,
  time        INTEGER NOT NULL,
  FOREIGN KEY (aid) REFERENCES account (aid),
  FOREIGN KEY (pid) REFERENCES protocol (pid)
) ;

CREATE INDEX ul_time_idx ON user_log(time);

--
-- User Data Collection table.
--
--  Record user text/number data
--
--  udc_id : user data collection id
--  key : indicate the record name
--  month : recorded month
--  number_entry : integer log entry
--  text_entry : string log entry
--

CREATE TABLE user_data_collection (
  udc_id       INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  key          TEXT NOT NULL,
  month        INTEGER NOT NULL,
  text_entry   TEXT,
  number_entry INTEGER
);

END TRANSACTION;
