#!/usr/bin/env python2.7

__author__  = "Jacek Kolezynski"
__version__ = "0.0.3"

import argparse
from datetime import datetime, timedelta
import math
import re
import subprocess
import sys


class Cert:
    def __init__(self, subject, issuer, content, expiry=None, position = 0, missing=False):
        self.subject = subject
        self.issuer = issuer
        self.content = content
        self.expiry = expiry
        self.children = []
        self.position = position
        self.missing = missing
    
    def add_child(self, child):
        self.children.append(child)
        

def extract_certs_as_strings(cert_file):
    certs = []
    with open(cert_file) as whole_cert:
        cert_started = False
        content = ''
        for line in whole_cert:
            if '-----BEGIN CERTIFICATE-----' in line:
                if not cert_started:
                    content += line
                    cert_started = True
                else:
                    print('Error, start cert found but already started')
                    sys.exit(1)
            elif '-----END CERTIFICATE-----' in line:
                if cert_started:
                    content += line
                    certs.append(content)
                    content = ''
                    cert_started = False
                else:
                    print('Error, cert end found without start')
                    sys.exit(1)
            elif cert_started:
                    content += line
        
        if cert_started:
            print('The file is corrupted')
            sys.exit(1)

    return certs


def create_certs(certs_contents):
    certs = []
    position = 1
    for content in certs_contents:
        certs.append(create_cert(content, position))
        position += 1
    return certs

def create_cert(cert_content, position):
    proc = subprocess.Popen(['openssl', 'x509', '-text'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, _ = proc.communicate(cert_content.encode())

    subject = ''
    issuer = ''
    date = None
    try:
        for line in out.decode().split('\n'):
            match = re.match("^\s*(\w*): .*CN ?= ?(.*)$", line)
            if match:
                if match.group(1) == 'Subject':
                    subject = match.group(2)
                elif match.group(1) == 'Issuer':
                    issuer = match.group(2)
            else:
                m = re.match("^\s*Not After\s?: (?P<date>.*)GMT$", line)
                if m:
                    date = datetime.strptime(m.group(1).strip(), '%b %d %H:%M:%S %Y')
    except:
        for line in out.split('\n'):
            match = re.match("^\s*(\w*): .*CN ?= ?(.*)$", line)
            if match:
                if match.group(1) == 'Subject':
                    subject = match.group(2)
                elif match.group(1) == 'Issuer':
                    issuer = match.group(2)
            else:
                m = re.match("^\s*Not After\s?: (?P<date>.*)GMT$", line)
                if m:
                    date = datetime.strptime(m.group(1).strip(), '%b %d %H:%M:%S %Y')

    return Cert(subject, issuer, cert_content, expiry=date, position=position)


def construct_tree(certs):
    roots_dir = {} # stores only root certs here
    issuers_dir = {c.subject : c for c in certs} 
    for c in certs:
        if c.subject in roots_dir:
            c = roots_dir[c.subject]
            # this is not self-signed cert, but was added temporarily to roots by other cert as missing parent
            c.missing = False
            del roots_dir[c.subject]   
        
        if c.subject == c.issuer:
            # this is self-signed cert, lets add it to roots
            roots_dir[c.issuer] = c
        else:
            # not self-signed cert
            if c.issuer in roots_dir:
                roots_dir[c.issuer].add_child(c)
            elif c.issuer in issuers_dir:
                issuers_dir[c.issuer].add_child(c)
            else:
                # this is not self signed cert, and has no parent in roots yet
                # so let's create temporary root and add it to roots
                missing_root = Cert(c.issuer, 'Unknown issuer', '', missing=True)
                roots_dir[c.issuer] = missing_root
                missing_root.add_child(c)

    return [r for r in roots_dir.values()]
